# THIS FILE IS GENERATED FROM macrosynergy SETUP.PY
short_version: str = '0.0.2'
version: str = '0.0.2'
full_version: str = '0.0.2'
git_revision: str = 'edd9cf47cb561ddda5fadaeaed107c766f88b1f2'
release: bool = True
if not release:
    version = full_version
